/* This file is part Tempura: Interval Temporal Logic interpreter.
 *  
 * Copyright (C) 1987-2019, 2021, 2023 Roger Hale, Ben Moszkowski,
 * Antonio Cau 
 *
 * Tempura is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Tempura is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Tempura.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Changes made by  AC: Antonio Cau
 *                 BCM: Ben Moszkowski
 *
 * $Id: globals.h 284 2017-10-08 08:00:31Z cau $
 */

#include "fns.h"

extern unsigned Ready;
extern unsigned Retry;
extern unsigned NextFlag;
extern unsigned int PrevFlag;
extern unsigned RedefFlag;
extern NODEPTR EmptyLoc;
extern unsigned Clock;
extern unsigned AlarmGone;
extern unsigned IsInitial;


extern NODEPTR Parseroot;
extern int yydebug;
extern char *yytext;
extern int yylineno;
extern int yyerrorcount;
extern YYLTYPE yylloc;
extern char Linetext[];
extern char *plinetext;
extern int yycolnum;


extern unsigned INTERACTIVE;
extern unsigned EOFILE;
extern unsigned INSCANSTR;



/* counter for top-level commands */
extern unsigned Counter;

/* Memory */
extern NODE GlobalMemory[];
extern unsigned LocalMemUsed;
extern NODEPTR TempLoc;
extern NODEPTR iRememberList;
extern unsigned CurrBit;
extern unsigned NextBit;

extern unsigned NoFraming;
extern unsigned Framing;


extern unsigned OneLevelMemFlag;
extern unsigned TimeStampCheckFlag;

/* environment */
extern ENVSTACK EnvStack[];
extern int EnvTop;
extern int EnvBase;

/* reduce-stack */
extern REDUCESTACK ReduceStack[];
extern REDUCESTACKPTR ReduceStackTop;
extern REDUCESTACKPTR ReduceStackMax;
extern NODEPTR LastCommand;

/* tracing and debugging */
extern jmp_buf *BreakPoint;
extern int TempuraLevel;
extern int BreakDepth;
extern unsigned Break;
extern unsigned InterruptSet;
extern int PassLimitCounter;
extern unsigned PassProgressFlag;
extern unsigned StateProgressFlag;
extern unsigned ForceOutput;
extern unsigned RunFlag;
/*extern int SyntaxErrorCount;*/

/* statistics */
extern unsigned StateNumber;
extern unsigned PassNumber;
extern unsigned TotalPassCount;
extern unsigned TotalReduceCount;
extern unsigned TotalReduceTimes;
extern unsigned StateReduceCount;
extern unsigned PassReduceCount;
extern unsigned TotalSuccessfulReduceCount;
extern unsigned TotalSuccessfulReduceTimes;
extern unsigned StateSuccessfulReduceCount;
extern unsigned PassSuccessfulReduceCount;
extern int ReduceDepth;
extern int MaxReduceDepth;

/* Static values (very commonly used) */
extern NODEPTR True, False;
extern NODEPTR Zero, One;
extern NODEPTR Nil, Empty, More, Skip, Ranlen, Infinite, StateNr, Random, RRandom, SRandom, SRRandom,  ZRandom, ZRRandom, FRandom, FRRandom;
extern NODEPTR EndOfFile;

/* Heap */
extern HEAP Heap;
extern unsigned HeapSize;
extern unsigned FreeNodes;
extern unsigned TempNodes;
extern NODEPTR FreeNodeList;
extern NODEPTR TempNodeList;

/* Symbol Table */
extern NODEPTR SymbolTableTop;
extern NODEPTR SymbolTableMax;
extern NODEPTR HashTable[];

/* Wait list for system calls */
extern NODEPTR SysWaitList;

/* System variables (user definable). */
extern int PassLimit;
extern int Verbosity;
extern int PrintLevel;
extern int DebugLevel;
extern unsigned StatePrintFlag;
extern unsigned RedundancyFlag;
extern unsigned CallByNameFlag;
extern unsigned AssignAheadFlag;
extern unsigned BreakIsAbortFlag;
extern unsigned HopChop;

/* To fix the path problem:
   extern char *Path;
   extern char *HelpDir;
*/
extern NODEPTR Path;
extern NODEPTR HelpDir;
extern NODEPTR InFile;
extern NODEPTR OldInFile;
extern NODEPTR OutFile;
extern FILE *Fpcur;
extern char *InFilename;

extern int TimeStep;

extern int64_t MaxRandLen;

extern int64_t RandSeed;
extern uint64_t TRandSeed;

extern int64_t StateNrVal;
extern unsigned StateNrValTime;
extern int64_t RandomVal;
extern unsigned RandomValTime;

/*extern int64_t RRandSeed;
  extern int TRRandSeed;*/
extern int64_t RRandomVal;
extern unsigned RRandomValTime;

/*extern int64_t SRandSeed;
  extern int TSRandSeed;*/
extern double SRandomVal;
extern unsigned SRandomValTime;

/*extern int64_t SRRandSeed;
  extern int TSRRandSeed;*/
extern double SRRandomVal;
extern unsigned SRRandomValTime;

extern gmp_randstate_t  Randstate;
extern mpz_t ZRandSeed;
extern mpz_t ZRandomMax;
extern mpz_t ZRandomVal;
extern mpz_t ZRRandomVal;
extern unsigned ZRandomValTime;
extern unsigned ZRRandomValTime;
extern mpfr_t FRandomVal;
extern mpfr_t FRRandomVal;
extern unsigned FRandomValTime;
extern unsigned FRRandomValTime;


/* floats */
extern int64_t Precision;
extern int FLRnd;
extern mpfr_prec_t  MPFRPrecision;
extern mpfr_rnd_t MPFRRnd;
extern int TMPFRRnd;
/* Input/Output */
/*extern NODEPTR AuxVal;*/
extern int CharsOutput;
extern FILE *InputFile;
/*extern char InputText[];*/
extern char *InputFilename;

/* Parser */
extern unsigned Token;
/*extern unsigned Recover;*/
extern char *Keyword[MAXTOKEN-MINTOKEN];

/* Reduction functions */
#define MAXREDUCETOKEN COMMA-MINTOKEN

extern PFI Reduces[MAXREDUCETOKEN + 1][3];
extern char *Reduc[MAXREDUCETOKEN + 1][3];
/*extern PFI ReduceStmt[];
extern PFI ReduceVal[];
extern PFI ReduceRef[];
*/
