/* This file is part Tempura: Interval Temporal Logic interpreter.
 *  
 * Copyright (C) 1987-2019, 2021, 2023 Roger Hale, Ben Moszkowski,
 * Antonio Cau 
 *
 * Tempura is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Tempura is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Tempura.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Changes made by  AC: Antonio Cau
 *                 BCM: Ben Moszkowski
 *
 * $Id: r-linear.c 284 2017-10-08 08:00:31Z cau $
 */


#include "system.h"



/*****************************************\
*                                         *
* tempura linear-time control statements  *
*                                         *
\*****************************************/


/* OR e1 e2 */
void reduce_OR (NODEPTR e)
{
    NODEPTR e1;
    e = reduce_val (e);
    if (!Ready) {
        e1 = reduce_val (node_lnk_1(e));
	        if (Ready) {
		    if (e1 == True)
		        e = True;
		    else {
		        Retry = T;
		        e = node_lnk_2(e);
		    }
	        } else {
		    e1 = reduce_val (node_lnk_2(e));
	            if (Ready) {
		        if (e1 == True)
		            e = True;
		        else {
		            Retry = T;
		            e = node_lnk_1(e);
		        }
		    }
		}
    }
    result (e);
}

/* iOR e1 e2 */
void reduce_iOR (NODEPTR e)
{
    NODEPTR e1;
    e = reduce_val (e);
    if (!Ready) {
        e1 = reduce_val (node_lnk_1(e));
	        if (Ready) {
		    if (e1 == True)
		        e = True;
		    else {
		        Retry = T;
		        e = node_lnk_2(e);
		    }
	        } else {
		    e1 = reduce_val (node_lnk_2(e));
	            if (Ready) {
		        if (e1 == True)
		            e = True;
		        else {
		            Retry = T;
		            e = node_lnk_1(e);
		        }
		    }
		}
    }
    result (e);
}

/* AND e1 e2 */
void reduce_AND (NODEPTR e)
{
    NODEPTR e1,e2;
    e1 = e2 = node3 (iAND, False, e, NULL);
    while (e1 != NULL) {
	while (node_type(node_lnk_2(e1)) == AND) {
	    node_lnk_3(e1) = node3 (iAND, False, node_lnk_2(node_lnk_2(e1)), node_lnk_3(e1));
	    node_lnk_2(e1) = node_lnk_1(node_lnk_2(e1)) ;
	}
	e1 = node_lnk_3(e1);
    }
    Retry = T;
    result (e2);
}

/* iAND ready e next */
void reduce_iAND (NODEPTR e)
{
    NODEPTR * cnode_ptr = &e,
        cnode,
        temp;
    unsigned allready = T;

    while ((cnode = *cnode_ptr) != NULL) {
        if (node_lnk_1(cnode) == True) 
	    cnode_ptr = &(node_lnk_3(cnode));
	else if (node_type(temp = node_lnk_2(cnode)) == iAND) {
	    NODEPTR *temp_ptr = &(node_lnk_2(cnode));
	    while (*temp_ptr != NULL)
	        temp_ptr = &(node_lnk_3((*temp_ptr)) );
            *temp_ptr = node_lnk_3(cnode);
	    node_lnk_1(cnode) = node_lnk_1(temp);
	    node_lnk_2(cnode) = node_lnk_2(temp);
	    node_lnk_3(cnode)  = node_lnk_3(temp);
	}
	else {
            if (node_type(node_lnk_2(cnode) = reduce_stmt (temp)) == BOOL) {
		if (node_lnk_2(cnode) == True) 
		    *cnode_ptr = node_lnk_3(cnode);
		else {
		    Ready = T;
		    result (False);
		    return;
		}
	    }
            else {
		if (Ready)
                    node_lnk_1(cnode) = True;
	        else
	            allready = F;
                cnode_ptr = &(node_lnk_3(cnode));
	    }
        }
    }
    if ((Ready = allready)) {
        for (cnode = e; cnode != NULL; cnode = node_lnk_3(cnode))
            node_lnk_1(cnode) = False;
    }
    else {
	cnode_ptr = &e;
	temp = NULL;
        while ((cnode = *cnode_ptr) != NULL) {
            if (node_lnk_1(cnode) == True) {
                cnode_ptr = &(node_lnk_3(cnode));
            }
            else {
                *cnode_ptr = node_lnk_3(cnode);
                node_lnk_3(cnode) = temp;
                temp = cnode;
	    }
	}
        *cnode_ptr = temp;
    }
    result (e == NULL ? True : node_lnk_3(e) == NULL ? node_lnk_2(e) : e);
}


/* IMPLIES e1 e2   */
/* added 04/09/2001 */
void reduce_IMPLIES (NODEPTR e)
{
  Retry = T;
  result (node3 (IF, node_lnk_1(e), node_lnk_2(e), True));
}

/* where e1 e2 */
void reduce_WHERE(NODEPTR e)
{
    Retry = T;
    result (conj (node_lnk_2(e), node_lnk_1(e)));
}

/* IF b e2 e3  */
/* added 06/09/2001 */
void reduce_IF_stmt(NODEPTR e)
{
    NODEPTR b = reduce_val (node_lnk_1(e));
    if (Ready) {
        type_check (b, BOOL);
        result (node_bval(b) ? node_lnk_2(e) : node_lnk_3(e));
        Retry = T;
    }
}

/* IF b e2 e3  */
/* Added 06/09/2001 */
void reduce_IF_val(NODEPTR e)
{
    NODEPTR b = reduce_val (node_lnk_1(e));
    if (Ready) {
        type_check (b, BOOL);
        result (node_bval(b) ? node_lnk_2(e) : node_lnk_3(e));
        Retry = T;
    }
}

/* iIF b e1 e2 */
/* Used by reduce_IF_val when test is a NEXT term. */
void reduce_iIF_val(NODEPTR e)
{
  NODEPTR e1,e2;
  Ready = T;
  e1 = reduce_val(node_lnk_2(e));
  if ( Ready ) {
      e1 = copy_val (e1);
  };
  if ( Ready ) {
    e2 = reduce_val(node_lnk_3(e));
    if ( Ready ) {
	e2 = copy_val (e2);
    };
  };
  if ( Ready )
    result (node1 (NEXT, node3 (IF, node_lnk_1(node_lnk_1(e)) ,e1,e2)));
}



/* CALL f elist */
void reduce_CALL(NODEPTR e)
{
    NODEPTR f = node_lnk_1(e);
    NODEPTR lambda;
    NODEPTR env;
    NODEPTR bnd, bnd_1;
    NODEPTR parm, parm1;
    NODEPTR elist = node_lnk_2(e);
    NODEPTR blk;
    NODEPTR base_loc;
    NODEPTR arg_loc;
    NODEPTR loc1;
    unsigned int history=1;
    unsigned int bnd_size = 0;
    /*unsigned int root_bnd_parm_call_level = 0;*/
    int call_level_offset, offset;
    unsigned has_offset = 0;
    /*unsigned  is_indirect  = 0;
    unsigned is_static = 0;
    unsigned is_lambda = 0;*/

    
    
    lambda = reduce_val (f);
    
    
    if (Ready) {

      /*
        newline;
      printf("----> call: calling ");
      output_expn(f, -1);
      newline;
      */
      
      
      if (node_type(lambda) == LAMBDA) {

        /*
        printf("call: lambda = ");
        output_expn(lambda, -1);
        newline;
        */
        /*type_check (lambda, LAMBDA);*/
        
        env = node_lnk_1(lambda);
        
                
        for (parm = node_env_idlist(env); (parm != NULL) && (elist != NULL);
             parm = node_next(parm), elist = node_lnk_2(elist)) {
          /*printf("call lambda: parm = %s level = %d\n",node_nm_name(parm),node_nm_level(parm));*/
            
            arg_loc = reduce_ref (node_lnk_1(elist));
            if (node_mem_history(arg_loc) > history) {
                history = node_mem_history(arg_loc);
            }
        }
        elist = node_lnk_2(e);
        blk = mem_block2 (STATIC,  node_env_size(env), history, NoFraming);
        base_loc = node_blk_base(blk);
        
        for (parm = node_env_idlist(env); (parm != NULL) && (elist != NULL);
             parm = node_next(parm), elist = node_lnk_2(elist)) {
          /*
          printf("call lambda: parm = %s level = %d\n",node_nm_name(parm),node_nm_level(parm));
          
          printf("call lambda:  evaluating ");
          output_expn(node_lnk_1(elist), -1);
          newline;
          
          show_scope();
          */
          /*
          for (parm1 = node_env_idlist(EnvStack[EnvTop].env);
               (parm1 != NULL); parm1 = node_next(parm1)) {
            node_nm_call_level_offset(parm1) = node_env_call_level_offset(EnvStack[EnvTop].env);
          }
          */
          /*
          show_scope();
          */
           
          
            arg_loc = reduce_ref (node_lnk_1(elist));
            
            if (Ready) {
              
                if (arg_loc == TempLoc) {
                  /*
                  printf("call lambda: temploc\n");
                  */
                  /*
                  show_scope();
                  */
                  put_mem1 (new_loc2 (base_loc + node_nm_off(parm), STATIC, 1, 0),
                            get_mem (arg_loc));
                  /*
                    printf("call lambda: with \n");
                    output_val(get_mem(arg_loc));
                    newline;
                  */
                  /*
                    printf("call lambda: lambda argyyy env_level= %d, env_call_level_offset = %d, bnd_size = %d, nm_level = %d \n",
                           node_env_level(EnvStack[EnvTop].env), node_env_call_level_offset(EnvStack[EnvTop].env), bnd_size, node_nm_level(parm));
                  */
                  /*
                    printf("call lambda: lambda new env_call_level_offset = %d\n", (node_env_level(EnvStack[EnvTop].env) + node_env_call_level_offset(EnvStack[EnvTop].env)+(bnd_size+1))-node_nm_level(parm));
                  */
                  if (node_type(get_mem(arg_loc)) == LAMBDA) {
                    /*
                      printf("call: got lambda\n");
                    */
                    call_level_offset = (node_env_level(EnvStack[EnvTop].env)+(bnd_size+1))-node_nm_level(parm);
                    /*
                      printf("call lambda: lambda arg call_level_offset=%d\n",call_level_offset);
                    */   
                    if (call_level_offset>0) {
                      if (call_level_offset > node_env_call_level_offset(env)) {
                        /* node_nm_call_level_offset(parm) = call_level_offset;*/
                        node_env_call_level_offset(env) = call_level_offset;
                      } else {
                        /* node_nm_call_level_offset(parm) = node_env_call_level_offset(env);*/
                      }
                    } else {
                      /* node_nm_call_level_offset(parm) = 0;*/
                    }
                  } else if (node_type(get_mem(arg_loc)) == iLAMBDA) {
                    call_level_offset = (node_env_level(EnvStack[EnvTop].env)+(bnd_size+1))-node_nm_level(parm);
                    /*  
                      printf("call lambda: ilambda arg call_level_offset=%d\n",call_level_offset);
                    */
                    if (call_level_offset>0) {
                      if (call_level_offset > node_env_call_level_offset(env)) {
                        /* node_nm_call_level_offset(parm) = call_level_offset;*/
                        node_env_call_level_offset(env) = call_level_offset;
                      } else {
                        /*  node_nm_call_level_offset(parm) = node_env_call_level_offset(env);*/
                      }
                    } else {
                      /* node_nm_call_level_offset(parm) = 0;*/
                    }
                  } else {
                    /* node_nm_call_level_offset(parm) = 0;*/
                  }
                  /*
                    printf("call lambda: parm call_level_offset=%d\n",node_nm_call_level_offset(parm));
                  */
                   
                }
                else if (node_type(parm) == STATIC && !is_static_mem (arg_loc)) {
		    NODEPTR val = reduce_val (node_lnk_1(elist));
                    /*
                      printf("call lambda: static\n");
                    */
		    if (Ready) {
                        put_mem1 (new_loc2 (base_loc + node_nm_off(parm),STATIC, 1, 0),val);
                    }
		    else {
		        return;
                    }
		}
                else {
                  /*
                  printf("call lambda: indirect\n");
                  */
                  loc1 = new_loc2 (base_loc + node_nm_off(parm), node_type(parm), node_mem_history(arg_loc), is_framed_mem(arg_loc));
                  make_indirect_mem (loc1,arg_loc);
                  
                }
            }
            else {
                return;
            }
        }
        
        
        assert0 ((elist == NULL) && (parm == NULL),
                "wrong number of parameters in function call");
        Retry = T;
        /*
        if (node_type(f) == STATIC || node_type(f) == STATE) {
          printf("call: name = %s, defined at level = %d\n",node_nm_name(f),node_nm_level(f));
        } else {
          printf("call: anonymous lambda \n");
        }
        */
        

        /* update the call level (offset?) of env and parm in
           node_env_idlist(env) here,
           maybe check call level of f if it is a name */

        /*
        printf("call lambda: node_env_call_level_offset = %d\n",
               node_env_call_level_offset(env));
        */
        if (node_env_call_level_offset(env)>0) {
          /*
            printf("call lambda: sanitize node_nm_call_level_offset env \n");
          */
          for (parm = node_env_idlist(env); (parm != NULL); parm = node_next(parm)) {
           
            node_nm_call_level_offset(parm) = node_env_call_level_offset(env);
          }
        }


        
        /*  
        printf("call: lambda nesting level = %d\n", bnd_size+1);
        printf("call: current level = %d\n", node_env_level(EnvStack[EnvTop].env));
        printf("call: current call level offset = %d\n", node_env_call_level_offset(EnvStack[EnvTop].env));
        */
        
       
          result (node3 (iCALL, f, bnd_node (node_lnk_1(lambda), blk, NULL), node_lnk_2(lambda)));
        
      } else if (node_type(lambda) == iLAMBDA) {

        /*
        printf("call: ilambda = ");
        output_expn(lambda, -1);
        newline;
        */
        /*type_check (lambda, LAMBDA);*/
        
        env = node_lnk_1(node_lnk_1(lambda));
        bnd = node_lnk_2(lambda);

        bnd_1 = bnd;
        while (bnd_1 != NULL) {
          bnd_size = bnd_size+1;
          bnd_1 = node_bnd_next(bnd_1);
        }
        
                
        for (parm = node_env_idlist(env); (parm != NULL) && (elist != NULL);
             parm = node_next(parm), elist = node_lnk_2(elist)) {
          /*printf("call: parm = %s level = %d\n",node_nm_name(parm),node_nm_level(parm));*/
            
            arg_loc = reduce_ref (node_lnk_1(elist));
            if (node_mem_history(arg_loc) > history) {
                history = node_mem_history(arg_loc);
            }
        }
        elist = node_lnk_2(e);
        blk = mem_block2 (STATIC,  node_env_size(env), history, NoFraming);
        base_loc = node_blk_base(blk);
        
        for (parm = node_env_idlist(env); (parm != NULL) && (elist != NULL);
             parm = node_next(parm), elist = node_lnk_2(elist)) {
          /*
          printf("call ilambda: parm = %s level = %d\n",node_nm_name(parm),node_nm_level(parm));
          printf("call ilambda:  evaluating ");
          output_expn(node_lnk_1(elist), -1);
          newline;
          */
          /*show_scope();*/
          /*
          for (parm1 = node_env_idlist(EnvStack[EnvTop].env);
               (parm1 != NULL); parm1 = node_next(parm1)) {
            node_nm_call_level_offset(parm1) = node_env_call_level_offset(EnvStack[EnvTop].env);
          }
          */
          /*
          show_scope();
          */
            arg_loc = reduce_ref (node_lnk_1(elist));
            
            if (Ready) {
                if (arg_loc == TempLoc) {
                  /*
                  printf("call ilambda: temploc\n");
                  */
                  
                    put_mem1 (new_loc2 (base_loc + node_nm_off(parm), STATIC, 1, 0),
                             get_mem (arg_loc));
                    /*printf("call ilambda: \n");
                    output_val(get_mem(arg_loc));
                    newline;*/
                    if (node_type(get_mem(arg_loc)) == LAMBDA) {
                      /*printf("call: got lambda\n");*/
                      call_level_offset = (node_env_level(EnvStack[EnvTop].env)+(bnd_size+1))-node_nm_level(parm);
                      /*
                      printf("call ilambda: lambda arg call_level_offset=%d\n",call_level_offset);
                      */
                      
                      if (call_level_offset>0) {
                        if (call_level_offset > node_env_call_level_offset(env)) {
                          /* node_nm_call_level_offset(parm) = call_level_offset;*/
                          node_env_call_level_offset(env) = call_level_offset;
                        } else {
                          /*  node_nm_call_level_offset(parm) = node_env_call_level_offset(env);*/
                        }
                      } else {
                        /*node_nm_call_level_offset(parm) = 0;*/
                      }
                    } else if (node_type(get_mem(arg_loc)) == iLAMBDA) {
                      call_level_offset = (node_env_level(EnvStack[EnvTop].env)+(bnd_size+1))-node_nm_level(parm);
                      /*
                      printf("call ilambda: ilambda arg call_level_offset=%d\n",call_level_offset);
                      */
                      if (call_level_offset>0) {
                        if (call_level_offset > node_env_call_level_offset(env)) {
                          /* node_nm_call_level_offset(parm) = call_level_offset;*/
                          node_env_call_level_offset(env) = call_level_offset;
                        } else {
                          /* node_nm_call_level_offset(parm) = node_env_call_level_offset(env);*/
                        }
                      } else {
                        /* node_nm_call_level_offset(parm) = 0;*/
                      }
                    } else
                    {
                      /* node_nm_call_level_offset(parm) = 0;*/
                    }
                    /*
                    printf("call ilambda: parm call_level_offset=%d\n",node_nm_call_level_offset(parm));
                    */
                }
                else if (node_type(parm) == STATIC && !is_static_mem (arg_loc)) {
		    NODEPTR val = reduce_val (node_lnk_1(elist));
                    /*
                    printf("call ilambda: static\n");
                    */
		    if (Ready) {
                        put_mem1 (new_loc2 (base_loc + node_nm_off(parm),STATIC, 1, 0),val);
                    }
		    else {
		        return;
                    }
		}
                else {
                  /*
                  printf("call ilambda: indirect\n");
                  */
                  loc1 = new_loc2 (base_loc + node_nm_off(parm), node_type(parm), node_mem_history(arg_loc), is_framed_mem(arg_loc));
                  make_indirect_mem (loc1,arg_loc);
                }
            }
            else {
                return;
            }
        }
        
        
        assert0 ((elist == NULL) && (parm == NULL),
                "wrong number of parameters in function call");
        Retry = T;
        /*
        if (node_type(f) == STATIC || node_type(f) == STATE) {
          printf("call: name = %s, defined at level = %d\n",node_nm_name(f),node_nm_level(f));
        } else {
          printf("call: anonymous lambda \n");
        }
        */
        

        /* update the call level (offset?) of env and parm in
           node_env_idlist(env) here,
           maybe check call level of f if it is a name */

        /*printf("call in ilambda: call_level_offset = %d, env= %d\n", call_level_offset, node_env_call_level_offset(env), );*/
        if (node_env_call_level_offset(env)>0) {
          /*
            printf("call ilambda: sanitize node_nm_call_level env \n");
          */
          for (parm = node_env_idlist(env); (parm != NULL); parm = node_next(parm)) {
            /*
            printf("call ilambda: old node_nm_call_level_offset = %d \n",node_nm_call_level_offset(parm));
            */
            node_nm_call_level_offset(parm) = node_env_call_level_offset(env);
          }
        }


        
        /* 
        printf("call ilambda: lambda nesting level = %d\n", bnd_size+1);
        printf("call ilambda: current level = %d\n", node_env_level(EnvStack[EnvTop].env));
        printf("call ilambda: current call level offset = %d\n", node_env_call_level_offset(EnvStack[EnvTop].env));
        */
        
        if (bnd != NULL) {
          /*printf("bnd is not null\n");*/
          bnd_1 = bnd;

          
          while (node_bnd_next(bnd_1) != NULL) {
            bnd_1 = node_bnd_next(bnd_1);
            if (node_env_call_level_offset(env) < node_env_call_level_offset(node_bnd_env(bnd_1))) {
              /*
              printf("call ilambda: env call_level_offset < bnd call_level_offset\n");
              */
              node_env_call_level_offset(env) = node_env_call_level_offset(node_bnd_env(bnd_1));
              for (parm = node_env_idlist(env); (parm != NULL); parm = node_next(parm)) {
                node_nm_call_level_offset(parm) = node_env_call_level_offset(env);
              }
            } else {
              if (node_env_call_level_offset(node_bnd_env(bnd_1)) < node_env_call_level_offset(env)) {
                /*
                printf("call ilambda: bnd call_level_offset < env call_level_offset\n");
                */
                node_env_call_level_offset(node_bnd_env(bnd_1)) = node_env_call_level_offset(env);
                for (parm = node_env_idlist(node_bnd_env(bnd_1)); (parm != NULL); parm = node_next(parm)) {
                  node_nm_call_level_offset(parm) = node_env_call_level_offset(env);
                }
              }
            }
          }
          if (node_env_call_level_offset(env) < node_env_call_level_offset(node_bnd_env(bnd_1))) {
            /*
            printf("call ilambda: env call_level_offset < bnd call_level_offset\n");
            */
            node_env_call_level_offset(env) = node_env_call_level_offset(node_bnd_env(bnd_1));
            for (parm = node_env_idlist(env); (parm != NULL); parm = node_next(parm)) {
              node_nm_call_level_offset(parm) = node_env_call_level_offset(env);
              }
          } else {
            if (node_env_call_level_offset(node_bnd_env(bnd_1)) < node_env_call_level_offset(env)) {
              /*
              printf("call ilambda: bnd call_level_offset < env call_level_offset\n");
              */
              node_env_call_level_offset(node_bnd_env(bnd_1)) = node_env_call_level_offset(env);
              for (parm = node_env_idlist(node_bnd_env(bnd_1)); (parm != NULL); parm = node_next(parm)) {
                  node_nm_call_level_offset(parm) = node_env_call_level_offset(env);
              }
            }
          }
          
                
          node_bnd_next(bnd_1) = bnd_node (node_lnk_1(node_lnk_1(lambda)), blk, NULL);

          result (node3 (iCALL, f, bnd, node_lnk_2(node_lnk_1(lambda))));
        } else {
          result (node3 (iCALL, f, bnd_node (node_lnk_1(node_lnk_1(lambda)), blk, NULL), node_lnk_2((node_lnk_1(lambda)))));
        }

      } else {
        assert1 ( (node_type(lambda) == LAMBDA) || (node_type(lambda) == iLAMBDA) ,
            "found expression of type %s when expecting lambda or ilambda",
             print_name (lambda));
      }
  }
}

/* iCALL env body baseloc */
void reduce_iCALL_stmt(NODEPTR e)
{
    NODEPTR bnd_list;
    int size = 0 ;

    /*
    printf("stmt: call level offset env = %d\n",node_env_call_level_offset(node_bnd_env(node_lnk_2(e))));
    */
    
    bnd_list = node_lnk_2(e);
    push_display (bnd_list);
    
    while (node_bnd_next(bnd_list) != NULL) {
      bnd_list = node_bnd_next(bnd_list);
      push_display(bnd_list);
      size = size+1;
    }
    /*
    printf("stmt: evaluating ");
    output_expn(node_lnk_3(e), -1);
    newline;
    printf("stmt: with scope\n");
    show_scope();
    */
    
    node_lnk_3(e) = reduce_stmt (node_lnk_3(e));

    pop_display ();
    
    while (size > 0) {
      pop_display();
      size = size-1;
    }

    if (node_type(node_lnk_3(e)) == LAMBDA) {
      /*printf("stmt: found lambda\n");*/
      /*node_lnk_3(node_lnk_3(e)) = node_lnk_2(e);*/
      node_lnk_3(e) = node2(iLAMBDA, node_lnk_3(e), node_lnk_2(e));
      
    } else if (node_type(node_lnk_3(e)) == iLAMBDA) {
      /*printf("stmt: found ilambda\n");*/
      node_lnk_3(e) = node2(iLAMBDA, node_lnk_1(node_lnk_3(e)), node_lnk_2(node_lnk_3(e)));
      /*node_lnk_3(e) = node2(iLAMBDA, node_lnk_1(node_lnk_3(e)), node_lnk_2(e));*/
      
      
    }
    
    if (node_type(node_lnk_3(e)) == BOOL) {
      result (node_lnk_3(e));
    }
/*    else if (node_type(node_lnk_3(e)) == iCALL)*/
/*        result (node_lnk_3(e));             */  /* Tail recursion */
}

/* iCALL env body baseloc */
void reduce_iCALL_val(NODEPTR e)
{
    NODEPTR val;
    NODEPTR bnd_list;
    int size = 0;

    /*
    printf("val: call level offset env = %d\n",node_env_call_level_offset(node_bnd_env(node_lnk_2(e))));
    */
    
    bnd_list = node_lnk_2(e);

    push_display (bnd_list);
    
    while (node_bnd_next(bnd_list) != NULL) {
      bnd_list = node_bnd_next(bnd_list);
      push_display(bnd_list);
      size = size+1;     
    }
    /*
    printf("val: evaluating ");
    output_expn(node_lnk_3(e), -1);
    newline;
    printf("val: with scope\n");
    show_scope();
    */
    
    val = reduce_val (node_lnk_3(e));
    
    pop_display ();
    
    while (size > 0) {
      pop_display();
      size = size-1;
    }
    /*
    printf("val: after evalation  scope is\n");
    show_scope();
    printf("val: Ready = %d\n",Ready);
    */
    
    if (Ready) {
      if (node_type(val) == LAMBDA) {
        
        /*val = copy_lambda(val);*/
        /*
        printf("val: found lambda\n");
        printf("val: old value\n");
        output_expn(val, -1);
        newline;
        */
        /*output_val_next(val);
          newline;*/
        
        /*node_lnk_3(val) = node_lnk_2(e);*/
        
        /*printf("new lambda\n");
        output_val(val);
        newline;*/
        /*output_expn( node_lnk_2(e), -1);
          newline;*/
        val = node2(iLAMBDA, val, node_lnk_2(e));
        /*
        printf("val: new value\n");
        output_expn(val, -1);
        newline;
        */
        /*newline;
        output_val(val);
        newline;
        output_val_next(val);
        newline;*/
        /*output_val_next(val);
          newline;*/
        
      } else if (node_type(val) == iLAMBDA) {
        /*
        printf("val: found ilambda\n");
        printf("val: call level offset env val = %d\n",node_env_call_level_offset(node_bnd_env(node_lnk_2(val))));
        printf("val: old value\n");
        output_expn( val, -1);
        newline;
        output_expn( node_lnk_2(e), -1);
        newline;
        */
        /*
        bnd_list = node_lnk_2(e);
        while (node_bnd_next(bnd_list) != NULL) {
           bnd_list = node_bnd_next(bnd_list);
        }
        node_bnd_next(bnd_list) = node_lnk_2(val);
        */
        val = node2(iLAMBDA, node_lnk_1(val), node_lnk_2(val));
        /*
        printf("val: new value\n");
        output_expn(val, -1);
        newline;
        */
        /* val = node2(iLAMBDA, node_lnk_1(val), node_lnk_2(e));*/
      }
      
      result (val);

    }
    
}

/* iCALL env body baseloc */
/* added memoryp 06/09/2001 */
void reduce_iCALL_ref(NODEPTR e)
{
    NODEPTR loc;
    NODEPTR val;
    NODEPTR bnd_list;
    int size = 0;

    /*
    printf("ref: call level offset env = %d\n",node_env_call_level_offset(node_bnd_env(node_lnk_2(e))));
    */
    
    bnd_list = node_lnk_2(e); 
    push_display (bnd_list);
    
    while (node_bnd_next(bnd_list) != NULL) {
      bnd_list = node_bnd_next(bnd_list); 
      push_display(bnd_list);
      size = size+1;
    }

    /*
    printf("ref: evaluating ");
    output_expn(node_lnk_3(e), -1);
    newline;
    printf("ref: with scope\n");
    show_scope();
    */
    
    loc = reduce_ref (node_lnk_3(e));
    
    pop_display ();
    while (size > 0) {
      pop_display();
      size = size-1;
    }

    /*
    printf("ref: after evalation  scope is\n");
    show_scope();
    printf("ref: Ready = %d\n",Ready);
    */
    
    if (Ready) {
      
      if ( mem_defined1 (loc) ) {
          val = get_mem(loc);
          
          if ( node_type(val) == LAMBDA) {
            /*
            printf("ref: found lambda\n");
            printf("ref: old value = ");
            output_expn(val, -1);
            newline;
            printf("ref: bnd = ");
            output_expn(node_lnk_2(e), -1);
            newline;
            */
            
            /*output_val(val);
              newline;*/
            
            /*val = copy_val(val);*/
            
            /*node_lnk_3(val) = node_lnk_2(e);*/
            
            /*printf("new lambda\n");
            output_val(val);
            newline;*/
            /*loc = make_ref(val);*/
            val = node2(iLAMBDA, val, node_lnk_2(e));
            /*
            printf("ref: new value = ");
            output_expn(val, -1);
            newline;
            */
            
            loc = make_ref(val);
          } else if ( node_type(val) == iLAMBDA) {
            /*
            printf("ref: found ilambda\n");
            printf("ref: call level offset env val = %d\n",node_env_call_level_offset(node_bnd_env(node_lnk_2(val))));
            printf("ref: old value = ");
            newline;
            output_expn(val, -1);
            newline;
            printf("ref: bnd = ");
            output_expn(node_lnk_2(e), -1);
            newline;
            */
            /*
            bnd_list = node_lnk_2(e);
            while (node_bnd_next(bnd_list) != NULL) {
              bnd_list = node_bnd_next(bnd_list);
            }
            node_bnd_next(bnd_list) = node_lnk_2(val);
            */
            val = node2(iLAMBDA, node_lnk_1(val), node_lnk_2(val));
            /*
            printf("ref: new value = ");
            output_expn(val, -1);
            newline;
            */
            
            /*val = node2(iLAMBDA, node_lnk_1(val), node_lnk_2(e));*/
            loc = make_ref(val);

          }
          
          ref_result (loc);
        } else {
        /*
        printf("ref: no mem_defined1\n");
        */
          ref_result (loc);
        }
    }
}

/* EXISTS env e w  (this is for "exists v<e:w") */
/* EXISTS env w */
void reduce_EXISTS(NODEPTR e)
{
    NODEPTR env = node_lnk_1(e);
    NODEPTR id;
    NODEPTR env_blk;
    NODEPTR env_base;
    

    if (node_lnk_2(e) != NULL) {
        Retry = T;
        result (node1 (NOT,
                    node3 (FORALL, env, node_lnk_2(e), node1 (NOT, node_lnk_3(e)))));
    }
    else {
        if (DebugLevel>2) {
            printf("Exists: history = %d\n",node_env_history(env));
            printf("Exists: framed = %d\n",node_env_framed(env));
        }
        env_base = node_blk_base(env_blk = allocate_mem1 (node_env_size(env),node_env_history(env))) ;
        for (id = node_env_idlist(env); id != NULL; id = node_next(id)) {
            new_loc2 (env_base + node_nm_off(id), node_type(id), node_env_history(env), node_env_framed(env));
            if (node_env_framed(env)) {
                if (iRememberList == NULL) {
                    iRememberList = node2(COMMA,id,NULL);
                }
                else {
                    iRememberList = node2(COMMA,id,iRememberList);
                }
            }
        }
        Retry = T;
                
        if (iRememberList == NULL) {
          result (node2 (iEXISTS, bnd_node (env, env_blk, NULL), node_lnk_3(e)));
        }
        else {
            if (node_lnk_3(e) != NULL) {
                if (node_type(node_lnk_3(e)) == EXISTS) {
                  result (node2 (iEXISTS, bnd_node (env, env_blk, NULL), node_lnk_3(e)));
                }
                else {
                  result (node2 (iEXISTS, bnd_node (env, env_blk, NULL), node2(AND,node_lnk_3(e),node1(ALWAYS,node1(iREMEMBER,node1(MAKELIST,iRememberList)))) ));
                    iRememberList = NULL;
                }
            }
            else {
              result (node2 (iEXISTS, bnd_node (env, env_blk, NULL), node_lnk_3(e)));
            }
        }
    }
}


/* iEXISTS idlist body baseloc */
void reduce_iEXISTS_stmt(NODEPTR e)
{
    push_env (node_lnk_1(e));
    node_lnk_2(e) = reduce_stmt (node_lnk_2(e));
    pop_env ();
    if (node_type(node_lnk_2(e)) == BOOL) 
        result (node_lnk_2(e));
}

/* iEXISTS idlist body baseloc */
void reduce_iEXISTS_val(NODEPTR e)
{
    NODEPTR val;
    push_env (node_lnk_1(e));
    val = reduce_val (node_lnk_2(e));
    pop_env ();
    if (Ready) {
	type_check (val, BOOL);
        result (val);
    }
}

/* FORALL id n expn */
void reduce_FORALL(NODEPTR e)
{
    NODEPTR blk;
    int nvall;
    NODEPTR env = node_lnk_1(e);
    NODEPTR expn = node_lnk_3(e);
    NODEPTR n = reduce_val (node_lnk_2(e));
    if (Ready) {
        type_check (n, INT);
        nvall = node_ival(n);
        e = True;
        for (nvall -= 1; nvall >= 0; nvall -= 1) {
            blk = local_mem (node_type(node_env_idlist(env)));
            put_mem1 (adj_loc1 (node_blk_base(blk)), int_node (nvall));
            e = node2 (AND, node2 (iEXISTS, bnd_node (env, blk, NULL), expn), e);
        }
        Retry = T;
        result (e);
    }
}

/* FORALLIN id list expn */
void reduce_FORALLIN(NODEPTR e)
{
  NODEPTR env = node_lnk_1(e);
  NODEPTR list = reduce_val (node_lnk_2(e));
  NODEPTR expn = node_lnk_3(e);
  NODEPTR loc;
  NODEPTR blk;
  int nvall;
   if (Ready) {
        static_check (node_env_idlist(env) );
        
        switch (node_type(list)) {
          case LIST:
            nvall = get_int (list_size (list));
            break;
          case STRING:
            nvall = string_size (list);
            break;
          default:
            error_break1 ("cannot index expression of type: %s",
            print_name (list));
        }
        e = True;
        for (nvall -= 1; nvall >= 0; nvall -= 1) {
          blk = local_mem (node_type(node_env_idlist(env)));
          loc = node_blk_base(blk);
          /*printf("nvall = %d\n",nvall);*/
          switch (node_type(list)) {
            case LIST:
              make_indirect_mem (loc, list_base (list) + offset2 (nvall,node_mem_history(list_base (list))));
              break;
             case STRING:
               put_mem1 (curr_loc (loc), substring (list, nvall, nvall));
               break;
             default:
               error_break1 ("cannot index expression of type: %s",
                 print_name (list));
          }
          e = node2 (AND, node2 (iEXISTS, bnd_node (env, blk, NULL), expn), e);
        }
        Retry = T;
        /*output_expn (e, 5);*/
        result (e);
   }
}

/* THE x : expn */
void reduce_THE(NODEPTR e)
{
    noop (e);
}

/* ALWAYS w */
void reduce_ALWAYS(NODEPTR e)
{
    always (reduce_stmt (node_lnk_1(e)));
}

/* SOMETIME w */
void reduce_SOMETIME(NODEPTR e)
{
    Retry = T;
    result (node3 (IF, node_lnk_1(e), True, node1 (NEXT, e)));
}

/* HALT w */
void reduce_HALT(NODEPTR e)
{
    NODEPTR e1;

    e1 = reduce_val (node_lnk_1(e));
    if (Ready) {
        type_check (e1, BOOL);
        if (e1 == True) {
            put_empty (True);
            result (True);
        }
        else
            put_empty (False);
    }
    else {
	if (empty_defined ()) {
	    Retry = T;
	    result (get_empty () == True ?
	            node_lnk_1(e) :
		    conjtf (e, node1 (NOT, node_lnk_1(e))));
	}
    }
}

/* INIT w */
void reduce_INIT(NODEPTR e)
{
    Retry = T;
    result (node2 (SEMI, conj (Empty, node_lnk_1(e)), True));
}

/* FIN w */
void reduce_FIN(NODEPTR e)
{
    if ((Ready = empty_defined ())) {
        if (get_empty () == True) {
            Retry = T;
            result (node_lnk_1(e));
        }
    }
}

/* KEEP w */
/* void reduce_KEEP(e)
NODEPTR e;
{
    keep (reduce_stmt (node_lnk_1(e)));
}
*/


/* WNEXT w */
void reduce_WNEXT(NODEPTR e)
{
    Ready = T;
    result (node_lnk_1(e));
}

/* NEXT w */
void reduce_NEXT_stmt(NODEPTR e)
{
    put_empty (False);
    Ready = T;
    result (node_lnk_1(e));
}

/* NEXT e */
void reduce_NEXT_val(NODEPTR e)
{
    NODEPTR val;
    put_empty (False);
    assert0 (!NextFlag, "two levels of next in expression");
    if ((PrevFlag>0) && (node_type(node_lnk_1(e)) == STATE || node_type(node_lnk_1(e)) == STATIC))
        assert2d (PrevFlag-1<=node_mem_history(mem_loc(node_lnk_1(e))), "too many levels of previous, requested %d only got history of %d",PrevFlag-1,node_mem_history(mem_loc(node_lnk_1(e))));
    NextFlag = T;
    val = reduce_val (node_lnk_1(e));
    NextFlag = F;
    if (Ready)
        result (copy_val (val));
    
}

/* NEXT e */
void reduce_NEXT_ref(NODEPTR e)
{
    NODEPTR loc;
    put_empty (False);
    assert0 (!NextFlag, "two levels of next in expression");
    if ((PrevFlag>0) && (node_type(node_lnk_1(e)) == STATE || node_type(node_lnk_1(e)) == STATIC)) {
        assert2d (PrevFlag-1<=node_mem_history(mem_loc(node_lnk_1(e))), "too many levels of previous, requested %d only got history of %d",PrevFlag-1,node_mem_history(mem_loc(node_lnk_1(e))));
    }
    NextFlag = T;
    loc = reduce_ref (node_lnk_1(e));
    NextFlag = F;
    if (Ready)
        ref_result (loc);
}


/* PREV w */
void reduce_PREV_stmt(NODEPTR e)
{
    error_break0("you can't use prev as a statement");
}

/* NPREV w */
void reduce_NPREV_stmt(NODEPTR e)
{
    error_break0("you can't use nprev as a statement");
}

/* PREV e */
void reduce_PREV_val(NODEPTR e)
{
    NODEPTR val;
        /*put_empty (False);*/
        /*if (DebugLevel>2)
          printf("reduce_PREV_val: history=%d\n",history(node_lnk_1(e)));*/
    if (node_type(node_lnk_1(e)) == STATE || node_type(node_lnk_1(e)) == STATIC) {
        if (NextFlag) {
            if (PrevFlag>0) {
                assert2d (PrevFlag-1<node_mem_history(mem_loc(node_lnk_1(e))), "too many levels of previous, requested %d only got history of %d",PrevFlag-1,node_mem_history(mem_loc(node_lnk_1(e))));
            }
        }
        else
            assert2d (PrevFlag<node_mem_history(mem_loc(node_lnk_1(e))), "too many levels of previous, requested %d only got history of %d",PrevFlag,node_mem_history(mem_loc(node_lnk_1(e))));
    }
    PrevFlag = PrevFlag +1;
    val = reduce_val (node_lnk_1(e));
    PrevFlag = PrevFlag -1; 
    if (Ready)
        result (copy_val (val));
       
}

/* NPREV e n*/
void reduce_NPREV_val(NODEPTR e)
{
    NODEPTR val;
    NODEPTR n = reduce_val (node_lnk_2(e));
    if (Ready) {
        type_check (n, INT);
        assert0(node_ival(n)>0,"level in nprev should be greater than zero");
            /*if (DebugLevel>2) {
            printf("reduce_NPREV_val: n=%d\n",node_ival(n));
            printf("reduce_NPREV_val: history=%d\n",history(node_lnk_1(e)));
            }*/
        if (node_type(node_lnk_1(e)) == STATE || node_type(node_lnk_1(e)) == STATIC) {
            if (NextFlag) {
                if (PrevFlag>0) {
                    assert2d (PrevFlag-1+node_ival(n)<=node_mem_history(mem_loc(node_lnk_1(e))), "too many levels of previous, requested %d only got history of %d",PrevFlag-1+node_ival(n),node_mem_history(mem_loc(node_lnk_1(e))));
                }
            }
            else
                assert2d (PrevFlag+node_ival(n)<=node_mem_history(mem_loc(node_lnk_1(e))), "too many levels of previous, requested %d only got history of %d",PrevFlag+node_ival(n),node_mem_history(mem_loc(node_lnk_1(e))));
        }
        PrevFlag = PrevFlag +node_ival(n);
        val = reduce_val (node_lnk_1(e));
        PrevFlag = PrevFlag -node_ival(n); 
        if (Ready)
            result (copy_val (val));
    }  
}

/* PREV e */
void reduce_PREV_ref(NODEPTR e)
{
    NODEPTR loc;
    
        /*put_empty (False);*/
        /*if (DebugLevel>2) {
        printf("reduce_PREV_ref: history=%d\n",history(node_lnk_1(e)));
        }*/
    if (node_type(node_lnk_1(e)) == STATE || node_type(node_lnk_1(e)) == STATIC) {
        if (NextFlag) {
            if (PrevFlag>0) {
                assert2d (PrevFlag-1<node_mem_history(mem_loc(node_lnk_1(e))), "too many levels of previous, requested %d only got history of %d",PrevFlag-1,node_mem_history(mem_loc(node_lnk_1(e))));
            }
        }
        else
            assert2d (PrevFlag<node_mem_history(mem_loc(node_lnk_1(e))), "too many levels of previous, requested %d only got history of %d",PrevFlag+1,node_mem_history(mem_loc(node_lnk_1(e))));
    }
    PrevFlag = PrevFlag +1;
    loc = reduce_ref (node_lnk_1(e));
    PrevFlag = PrevFlag -1;
    if (Ready)
        ref_result (loc);
}

/* NPREV e n*/
void reduce_NPREV_ref(NODEPTR e)
{
    NODEPTR loc;
    NODEPTR n = reduce_val (node_lnk_2(e));
    if (Ready) {
        type_check (n, INT);
        assert0(node_ival(n)>0,"level in nprev should be greater than zero");
            /*if (DebugLevel>2) {
            printf("reduce_NPREV_ref: NextFlag=%d\n",NextFlag);
            printf("reduce_NPREV_ref: PrevFlag=%d\n",PrevFlag);
            printf("reduce_NPREV_ref: n=%d\n",node_ival(n));
            printf("reduce_NPREV_ref: history=%d\n",history(node_lnk_1(e)));
            }*/
        if (node_type(node_lnk_1(e)) == STATE || node_type(node_lnk_1(e)) == STATIC) {
            if (NextFlag) {
                if (PrevFlag>0) {
                    assert2d (PrevFlag-1+node_ival(n)<=node_mem_history(mem_loc(node_lnk_1(e))), "too many levels of previous, requested %d only got history of %d",PrevFlag-1+node_ival(n),node_mem_history(mem_loc(node_lnk_1(e))));
                }
            }
            else
                assert2d (PrevFlag+node_ival(n)<=node_mem_history(mem_loc(node_lnk_1(e))), "too many levels of previous, requested %d only got history of %d",PrevFlag+node_ival(n),node_mem_history(mem_loc(node_lnk_1(e))));
        }
        PrevFlag = PrevFlag +node_ival(n);
        loc = reduce_ref (node_lnk_1(e));
        PrevFlag = PrevFlag -node_ival(n);
        if (Ready)
            ref_result (loc);
    }
}

/* LEN e */
void reduce_LEN(NODEPTR e)
{
    NODEPTR e1 = reduce_val (node_lnk_1(e));
    if (Ready) {
        type_check (e1, INT);
        Retry = T;
        result (node1 (iLEN, int_node (node_ival(e1))));
    }
}

/* iLEN c */
void reduce_iLEN(NODEPTR e)
{
    NODEPTR c = node_lnk_1(e);
    if (node_ival(c) == 0) {
        put_empty (True);
        result (True);
    }
    else {
        node_ival(c) -= 1;
        put_empty (False);
    }
    Ready = T;
}

void reduce_RANLEN(NODEPTR e)
{
  Retry = T;
      /* result (node1 (iLEN, int_node ((MaxRandLen<0 ? 0 : i4_uniform_0i(&TRandSeed) % (MaxRandLen+1)))));*/
      /* result (node1 (iLEN, int_node ( (MaxRandLen<0 ? 0 : genrand64_int63() %  (MaxRandLen+1) ) ) ) );*/
  result (node1 (iLEN, int_node ( (MaxRandLen<0 ? 0 : xorshift128plus()  %  (MaxRandLen+1) ) ) ) );
}

void reduce_INFINITE(NODEPTR e)
{
  Retry = T;
    result (node1 (SOMETIME, False));
}

/* PROCESS e */
void reduce_PROCESS(NODEPTR e)
{
    Retry = T;
    result (RedundancyFlag ?
            node3 (iPROCESS, node_lnk_1(e), False, local_mem (STATE)) :
            node_lnk_1(e));
}

/* iPROCESS e ready empty_loc */
void reduce_iPROCESS(NODEPTR e)
{
    NODEPTR local_empty, global_empty;
    NODEPTR empty_loc = node_blk_base(node_lnk_3(e));
    
    if (!(Ready =node_bval(node_lnk_2(e)) )) {
        push_empty (empty_loc);
        node_lnk_1(e) = reduce_pgm (node_lnk_1(e));
        node_lnk_2(e) = Ready ? True : False;
        pop_empty;
    }
    if (Ready) {
        if ((Ready = empty_defined ())) {
            global_empty = get_empty ();
            local_empty = get_mem (curr_loc (empty_loc));
            assert0 (!(local_empty == False && global_empty == True),
                    "calling process terminated before subprocess");
            if (local_empty == True || node_type(node_lnk_1(e)) == BOOL)
                result (node_lnk_1(e));
            else 
                node_lnk_2(e) = False;
        }
    }
}

/* PREFIX e */
void reduce_PREFIX(NODEPTR e)
{
/*    assert0 (!AssignAheadFlag,*/
/*            "prefix operator requires \"assign_ahead\" to be false");*/
    Retry = T;
    result (node3 (iPREFIX, node_lnk_1(e), False, local_mem (STATE)));
}

/* iPREFIX e ready empty_loc */
void reduce_iPREFIX(NODEPTR e)
{
    NODEPTR local_empty, global_empty;
    NODEPTR empty_loc = node_blk_base(node_lnk_3(e));
    push_assign_ahead;

    if (!(Ready = node_bval(node_lnk_2(e)) )) {
        push_empty (empty_loc);
        node_lnk_1(e) = reduce_stmt (node_lnk_1(e));
        node_lnk_2(e) = Ready ? True : False;
        pop_empty;
    }
    if (Ready) {
        if ((Ready = empty_defined ())) {
            global_empty = get_empty ();
            local_empty = mem_defined1 (empty_loc = curr_loc (empty_loc)) ?
                         get_mem(empty_loc) :
                         global_empty;
            assert0 (local_empty == False || global_empty == True,
                    "prefix terminated before enclosing interval");
            if (global_empty == True)
                result (True);
            else if (node_type(node_lnk_1(e)) == BOOL)
	        result (node_lnk_1(e));
	    else
                node_lnk_2(e) = False;
        }
    }
    pop_assign_ahead;
}

